# -*- coding: utf-8 -*-


import asyncio

from contextlib import contextmanager


async def wait_until_cancelled(*, propagate=True, loop=None):
    """Wait until the calling task is canceled.

    **Note**: this function is a coroutine.

    When using a context manager to complete one or more background tasks, it's
    common to have the "main" task block until something cancels it (e.g. a
    SIGINT/CTRL-C handler).

    It's also convenient in tests that verify the behavior of cancellation to
    need to spawn a background task that waits forever.

    This leads to the not-so-idiomatic:

    .. code-block:: python

       await asyncio.Future()

    This of often wrapped in a helper function to make the call more readable.
    Instead of propagating multiple variants of this, it should be placed in a
    library that everybody can import.

    :param loop: Loop in which the coroutine will block.  Defaults to the
     current event loop.

    """

    loop = loop or asyncio.get_event_loop()

    if propagate:
        await asyncio.Future(loop=loop)
    else:
        try:
            await asyncio.Future(loop=loop)
        except asyncio.CancelledError:
            pass


async def cancel(task, loop=None):
    """Cancel a task and wait until it's done.

    **Note**: this function is a coroutine.

    Canceling a child task and returning without waiting for the child task to
    complete is a common cause of "event loop closed" ``RuntimeError``
    exceptions, especially during program shutdown.  Therefore, this becomes a
    common pattern:

    .. code-block:: python

        task.cancel()
        await asyncio.wait({task})

    However, if the parent task itself is also canceled, then the
    ``asyncio.wait()`` call will be interrupted and the child task will still
    not complete.  To solve this, we must also manage to trap the
    ``asyncio.CancelledError`` exception and call ``asyncio.wait({task})``
    again and properly re-raise the ``asyncio.CancelledError`` exception.  For
    example:


    .. code-block:: python

        task.cancel()
        try:
            await asyncio.wait({task})
        except asyncio.CancelledError:
            await asyncio.wait({task})
            raise

    This is not trivial and must be done so many times in a program that
    cancels tasks that it merits a replacement API for ``task.cancel()``.

    :param task: The ``asyncio.Task`` object to cancel.
    :param loop: The event loop to use for awaiting.  Defaults to the current
     event loop.

    .. versionadded: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    task.cancel()
    try:
        await asyncio.wait({task}, loop=loop)
    except asyncio.CancelledError:
        await asyncio.wait({task}, loop=loop)
        raise


async def cancel_all(tasks, loop=None):
    """Cancel a set of tasks and wait until they're done.

    **Note**: this function is a coroutine.

    Canceling a set of child tasks and returning without waiting for the child
    task to complete is a common cause of "event loop closed" ``RuntimeError``
    exceptions, especially during shutdown of servers with one ore more task
    per connection.  Therefore, this becomes a common pattern:

    .. code-block:: python

        for task in tasks:
            task.cancel()
        await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)

    However, if the parent task itself is also canceled, then the
    ``asyncio.wait()`` call will be interrupted and one or more of the child
    tasks will still not complete.  To solve this, we must also manage to trap
    the ``asyncio.CancelledError`` exception and call ``asyncio.wait(tasks)``
    again and properly re-raise the ``asyncio.CancelledError`` exception.  For
    example:

    .. code-block:: python

        for task in tasks:
            task.cancel()
        try:
            await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        except asyncio.CancelledError:
            await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
            raise

    This is not trivial and must be done so many server programs that cancels
    tasks that it merits a helper.

    :param tasks: The set of ``asyncio.Task`` objects to cancel.
    :param loop: The event loop to use for awaiting.  Defaults to the current
     event loop.

    .. versionadded: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    for task in tasks:
        task.cancel()
    try:
        await asyncio.wait(tasks, loop=loop)
    except asyncio.CancelledError:
        await asyncio.wait(tasks, loop=loop)
        raise


async def follow_through(task, loop=None):
    """Wait for a task to complete (even if canceled while waiting).

    **Note**: this function is a coroutine.

    Not propagating cancellation to a child task and returning without waiting
    for the child task to complete is a common cause of "event loop closed"
    ``RuntimeError`` exceptions, especially during program shutdown.
    Therefore, this becomes a common pattern:

    .. code-block:: python

        try:
            await asyncio.wait({task})
        except asyncio.CancelledError:
            task.cancel()
            await asyncio.wait({task})
            raise
        return task.result()

    This is not trivial and must be done so many times in a program that spawns
    child tasks that it merits a helper method.

    :param task: The ``asyncio.Task`` object to see through to completion.
    :param loop: The event loop to use for awaiting.  Defaults to the current
     event loop.

    .. versionadded: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    try:
        await asyncio.wait({task}, loop=loop)
    except asyncio.CancelledError:
        await cancel(task, loop=loop)
        raise
    return task.result()


def run_until_complete(coro, loop=None):
    """Run a task through to completion.

    The ``.run_until_complete()`` method on asyncio event loop objects doesn't
    finish tasks when it receives a SIGINT/CTRL-C.  The method simply raises a
    ``KeyboardInterrupt`` exception and this usually results in warnings about
    unfinished tasks plus some "event loop closed" ``RuntimeError`` exceptions
    in pending tasks.

    This is a really annoying default behavior and this function aims at
    replacing that behavior with something that ensures the task actually runs
    through to completion.  When the ``KeyboardInterrupt`` exception is caught,
    the task is canceled and resumed to give it a chance to clean up properly.

    .. versionadded: 0.4

    """

    loop = loop or asyncio.get_event_loop()
    task = loop.create_task(coro)
    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        task.cancel()
        try:
            loop.run_until_complete(task)
        except asyncio.CancelledError:
            return None
    return task.result()



@contextmanager
def reader(fd, callback, loop=None):
    """Register a low-level reader for a file descriptor.

    **Note**: the proactor event loop does not support readers.

    It's an easy mistake to forget a call to ``.remove_reader()`` and end up
    receiving extra I/O you were not interested in.  This context manager makes
    sure you never forget.

    :param fd: File descriptor to watch for read events.
    :param callback: Called when the file descriptor is ready to read from.
     See asyncio's documentation on ``loop.add_reader()`` for details.
    :param loop: Loop in which the watch will be registered.  Defaults to the
     current event loop.

    This context manager yields nothing.

    .. versionadded: 0.4

    """

    loop = loop or asyncio.get_event_loop()

    loop.add_reader(fd, callback)
    try:
        yield
    finally:
        loop.remove_reader(fd)


from .mempipe import mempipe
from .monkey import monkey_patch
from .posix import UnixSocketServer
from .stack import AsyncExitStack, EnsureDone
from .tcp import TCPServer, tcp_server
from ._udp import udp_server
from .testing import mock_subprocess
from .ctrlc import handle_ctrlc
from ._pool import PoolClosed, TaskPool


__all__ = [
    'AsyncExitStack',
    'cancel',
    'cancel_all',
    'follow_through',
    'handle_ctrlc',
    'mempipe',
    'mock_subprocess',
    'monkey_patch',
    'PoolClosed',
    'reader',
    'run_until_complete',
    'TaskPool',
    'TCPServer',
    'tcp_server',
    'udp_server',
    'UnixSocketServer',
    'wait_until_cancelled',
]



# -*- coding: utf-8 -*-


import asyncio
import logging

from aiotk import cancel, cancel_all, EnsureDone


class PoolClosed(Exception):
    pass


class TaskPool:
    """Collect background tasks.

    This is useful for server programs that wish to spawn one task per
    connection.  It removes the need to implement the redundant task management
    code required to close tasks as they complete and the need to implement
    graceful shutdown (with cancellation) to clean up when shutting down the
    server.

    """

    def __init__(self, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._lock = asyncio.Lock(loop=self._loop)
        self._cond = asyncio.Condition(self._lock, loop=self._loop)
        self._pool = set()
        self._task = None
        self._done = False
        self._idle = asyncio.Condition(self._lock, loop=self._loop)
        self._busy = asyncio.Condition(self._lock, loop=self._loop)

    async def spawn(self, fn, *args, **kwds):
        """Add a new task to collect.

        The pool is designed to keep memory usage as low as possible by
        collecting tasks as soon as possible; it keeps no references to the
        task once it has completed.  If you want to get the task's result, you
        can keep a reference to the ``asyncio.Task`` object that is returned.

        :param fn: The coroutine function to invoke.
        :param args: Positional arguments to pass to the coroutine function.
        :param kwds: Keyword arguments to pass to the coroutine function.
        :return: An ``asyncio.Task`` object.

        """

        async with self._lock:

            # Ensure we don't spawn tasks after ``.close()`` is called.
            if self._done:
                raise PoolClosed()

            # Spawn the task.
            task = self._loop.create_task(fn(*args, **kwds))

            # Add the task.
            self._pool.add(task)

            # Signal the background task to update its wait list.
            self._cond.notify()

        return task

    async def wait_idle(self):
        """Block until all tasks have completed.

        This is useful to perform fan-in: spawn a set of tasks and then wait
        for all of them to complete.

        """

        async with self._lock:
            if self._done:
                raise PoolClosed()
            await self._idle.wait()

    async def wait_busy(self):
        """Block until the task is known to be waiting for tasks to complete.

        This is mostly useful for testing purposes (it avoids flaky behavior
        due to timing and makes the code paths more predictable).

        """
        async with self._lock:
            if self._done:
                raise PoolClosed()
            await self._busy.wait()

    async def __aenter__(self):
        """Perform the full startup sequence."""

        self.start()
        await self.wait_started()
        return self

    async def __aexit__(self, *args):
        """Perform the full shutdown sequence.


        This unblocks once all tasks in the pool have completed and that the
        collection task has completed.

        """

        self.close()
        await self.wait_closed()

    def start(self):
        """Start the collection task."""
        assert self._task is None
        self._task = self._loop.create_task(self._collect())

    async def wait_started(self):
        """Wait until the collection task has started."""
        assert self._task is not None
        async with self._lock:
            await self._idle.wait()

    def close(self):
        """Start the shutdown sequence."""

        assert self._task is not None

        # NO-OP: there's nothing we can do with a synchronous ``.close()``.

    async def wait_closed(self):
        """Wait until the shutdown sequence has completed.

        This unblocks once all tasks in the pool have completed and that the
        collection task has completed.

        """

        # Signal all tasks to start the shutdown sequence.
        async with self._lock:
            self._done = True
            self._cond.notify()

        # NOTE: don't hold the lock here else we'll prevent the shutdown from
        #       happening.
        if self._pool:
            await cancel_all(self._pool, loop=self._loop)
        await cancel(self._task)

    async def _collect(self):
        """Collect tasks that complete in the background."""

        # Main collection loop.
        while True:

            # Block until we have at least one task in the pool or until we are
            # told to shut down, which ever comes first.
            while not self._done and not self._pool:
                async with self._lock:
                    self._idle.notify()
                    await self._cond.wait()

            # Build a wait set containing all the tasks, plus a watch for
            # new tasks.
            async with self._lock:
                pending = set(task for task in self._pool)
                async with EnsureDone(self._cond.wait(),
                                      loop=self._loop) as watch:
                    pending.add(watch)

                    # NOTE: `self._cond.wait()` releases the lock until we're
                    #       notified, so we don't hold the lock while we block
                    #       on `asyncio.wait()`.
                    self._busy.notify()
                    done, pending = await asyncio.wait(
                        pending, return_when=asyncio.FIRST_COMPLETED,
                    )
                    for task in done:
                        if task is watch:
                            continue
                        self._pool.remove(task)

                        # Log exceptions for crashed tasks.
                        try:
                            task.result()
                        except asyncio.CancelledError:
                            pass
                        except Exception:
                            logging.exception('Task crashed!')



# -*- coding: utf-8 -*-


import asyncio
import sys

from aiotk import cancel, follow_through


class EnsureDone:
    """Ensure a background task completes before leaving a code block.

    This is mostly useful for spawning multiple tasks in the background using
    :py:class:`~AsyncExitStack`, but it can also be used directly for single
    tasks.

    .. source-code: python

    .. versionadded: 0.3

    """

    def __init__(self, coro, cancel=True, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._coro = coro
        self._task = None
        self._cancel = cancel

    async def __aenter__(self):
        """Spawn the background task."""

        self._task = self._loop.create_task(self._coro)
        return self._task

    async def __aexit__(self, *args):
        """Wait until the task completes.

        If ``cancel=True`` was passed to the constructor, the task is cancelled
        before being awaited.

        """
        if self._cancel:
            await cancel(self._task, loop=self._loop)
        else:
            await follow_through(self._task, loop=self._loop)


class AsyncExitStack(object):
    """Rollback stack for asynchronous context managers.

    This context manager provides the following features over direct use of
    ``contextlib.ExitStack``:

    - supports asynchronous context managers (in addition to regular context
      managers).

    .. versionadded: 0.2

    """

    def __init__(self):
        self._stack = []

    async def __aenter__(self):
        """No-op."""
        assert self._stack == []
        return self

    async def __aexit__(self, etype, exc, tb):
        """Pop all context managers from the rollback stack.

        :param context: context manager or asynchronous context manager.
        :returns: True if one of the context managers in the stack suppressed
         the exception by returning a ``True`` value from its ``__exit__`` or
         ``__aexit__`` method.
        :exception: If any context manager in the stack raises an exception
         from its ``__exit__`` or ``__aexit__`` method, this exception will
         replace the original exception for context managers lower in the stack
         and will eventually be propagated to the caller.

        """
        changed = False
        for context in reversed(self._stack):
            suppress = False
            try:
                if hasattr(context, '__aexit__'):
                    suppress = await context.__aexit__(etype, exc, tb)
                if hasattr(context, '__exit__'):
                    suppress = context.__exit__(etype, exc, tb)
                if suppress:
                    etype, exc, tb = (None, None, None)
                    changed = True
            except:
                etype, exc, tb = sys.exc_info()
                changed = True
        if changed:
            if (etype, exc, tb) == (None, None, None):
                return True
            raise exc

    async def enter_context(self, context):
        """Push an (asynchronous) context manager onto the rollback stack.

        :param context: context manager or asynchronous context manager.
        :return: The return value of the context manager's ``__enter__`` or
         ``__aenter__`` method.

        """
        if hasattr(context, '__aenter__'):
            r = await context.__aenter__()
        else:
            r = context.__enter__()
        self._stack.append(context)
        return r



# -*- coding: utf-8 -*-


from setuptools import find_packages, setup


def readfile(path):
    with open(path, 'rb') as stream:
        return stream.read().decode('utf-8')


version = readfile('src/aiotk/version.txt').strip()
readme = readfile('README.rst')


setup(
    name='aiotk',
    maintainer='Andre Caron',
    maintainer_email='andre.l.caron@gmail.com',
    version=version,
    url='https://github.com/AndreLouisCaron/aiotk',
    packages=find_packages(where='src'),
    package_dir={
        '': 'src',
    },
    package_data={
        'aiotk': [
            'version.txt',
        ],
    },
    long_description=readme,
)



# -*- coding: utf-8 -*-


import asyncio
import socket

from contextlib import contextmanager
from functools import partial
from .stack import AsyncExitStack, EnsureDone


@contextmanager
def udp_socket(host, port):
    """Create, bind and cleanup a UDP socket."""

    s = socket.socket(
        socket.AF_INET,
        socket.SOCK_DGRAM,
        socket.IPPROTO_UDP,
    )
    try:
        s.bind((host, port))
        yield s
    finally:
        s.close()


def udp_reader(socket, iqueue, size):
    """Read one or more packets from an UDP socket."""

    data, peer = socket.recvfrom(size)
    iqueue.put_nowait((peer, data))


async def udp_writer(socket, oqueue):
    """Forward packets to the UDP socket."""

    while True:
        peer, data = await oqueue.get()
        try:
            socket.sendto(data, peer)
        finally:
            oqueue.task_done()


async def udp_server(host, port, service, loop=None):
    """Simple UDP-based service.

    The only examples for UDP in asyncio documentation use protocols, the
    callback-based APIs and are a bit confusing (connection made?  connection
    lost?).

    This helper method tries to turn the low-level UDP socket support into a
    stream-based API.  You pass in a coroutine function to which a pair of
    queues will be passed.  From there, you can use async/await syntax to send
    and receive packets.

    :param host: Network interface on which to bind.
    :param port: Port number on which to bind.
    :service: coroutine that will perform logic.
    :param loop: Loop in which the service will run.

    The ``service`` coroutine should have the following signature:

    .. code-block:: python

       async def my_udp_service(*, iqueue, oqueue, loop, **kwds):
           pass

    The ``iqueue`` and ``oqueue`` parameters are ``asyncio.Queue`` objects that
    the coroutine can use to read from and write to, respectively.

    .. versionadded: 0.4

    """

    # Circular imports (yuk, fixme)!
    from . import reader

    loop = loop or asyncio.get_event_loop()

    async with AsyncExitStack() as stack:

        # Create & bind the socket.
        socket = await stack.enter_context(udp_socket(host, port))

        # Pair of queues through which packets will travel.
        iqueue = asyncio.Queue(loop=loop)
        oqueue = asyncio.Queue(loop=loop)

        # Forward packets from the queue to the socket.
        await stack.enter_context(EnsureDone(
            udp_writer(socket=socket, oqueue=oqueue)
        ))

        try:
            # Forward packets from the socket to the queue.
            #
            # NOTE: if we want to be able to wait until all packets are sent
            #       using ``oqueue.join()``, we need to stop reading before we
            #       do that, which is why the ``reader()`` context manager is
            #       registered after (cleanup is LIFO).
            await stack.enter_context(reader(
                socket,
                partial(
                    udp_reader, socket=socket,
                    iqueue=iqueue, size=2048,
                ),
            ))

            # Respond in the foreground until we're cancelled.
            return await service(iqueue=iqueue, oqueue=oqueue, loop=loop)
        finally:
            # Make sure to "flush" the output queue before leaving.
            await oqueue.join()



# -*- coding: utf-8 -*-


import asyncio
import contextlib
import signal


@contextlib.contextmanager
def handle_ctrlc(f, loop=None):
    """Context manager that schedules a callback when SIGINT is received.

    :param f: future to fulfill when SIGINT is received.  Will only be set
     once even if SIGINT is received multiple times.
    :param loop: event loop (defaults to global event loop).

    """

    loop = loop or asyncio.get_event_loop()

    def handler():
        if not f.done():
            f.set_result(None)

    loop.add_signal_handler(signal.SIGINT, handler)
    try:
        yield
    finally:
        loop.remove_signal_handler(signal.SIGINT)
